/* File: main.c
 * Desc: Flashtest stub of Zaphod 1
 * Auth: Woody
 * Note: 
 */

#include <termios.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>

#include <fcntl.h>
#include <sys/signal.h>
#include <sys/types.h>

#include <znodelist.h>
#include <afp.h>
#include <config.h>
#include <portfw.h>

#define BAUDRATE B19200					//38400
#define MODEMDEVICE "/dev/ttyS0"
#define _POSIX_SOURCE 1					/* POSIX compliant source */
#define MAX(a,b) ((a>b)?a:b)

#define ARE_U_THERE 0xC1
#define	NODE_STATUS 0xD9
#define	NODE_CONFIG 0xEB
#define BAD_LINK_PACKET 1

#define STALE_PACKET_TO	1
#define REQUEST_TO	20

#define	REPLY_TIMEOUT -1
#define	GOT_REPLY	-2
#define	REPLY_NO_DATA -3

#define	EEPROM_SIZE	360
#define deb(x) x

CONFIGURATION *Tconfigp_ptr;
//Z_NODE *ZN_anchor = 0;
extern int errno;

char *mytime ();
int node_link (int link_fd, int opcode, char* buff);
int	wait_response(int link_fd, int opcode, char* buff);
int	read_eeprom(int link_fd);
int	get_all_nodes(int link_fd);


char *eeprom_ptr;

char* config_ptr = NULL;

int I_flushflag = 0;

/*********************************************************************/

int 
zflash (void)
{
	int I_sockfd;
	struct sockaddr_in serv_addr, cli_addr;
	void *Vp_buff;

	int link_fd, c, res;
	struct termios oldtio, newtio;


	system ("clear");

#if 0
	if (!(eeprom_ptr = (char *) malloc (EEPROM_SIZE)))
	{
		disp_label ("Unable to allocate memory for eeprom data");
		return (1);
	}
#endif

	/* Initialize serial comm system */
	disp_label ("Initializing facility link system");

	/* open the device to be non-blocking (read will return immediatly) */
	link_fd = open (MODEMDEVICE, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (link_fd < 0)
	{
		perror ("facility link");
		return (1);
	}

	tcgetattr (link_fd, &oldtio);			/* save current port settings */
	/* set new port settings for non-canonical input processing */
	bzero (&newtio, sizeof (newtio));

	newtio.c_cflag = BAUDRATE | PARENB | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNBRK | INPCK | ICRNL;
	newtio.c_oflag = 0;
	newtio.c_lflag = 0;
	newtio.c_cc[VMIN] = 1;				// wait for up to 32 chars
	newtio.c_cc[VTIME] = 1;				// * 100 ms

	tcflush (link_fd, TCIFLUSH);
	tcsetattr (link_fd, TCSANOW, &newtio);

#if 0
	if (read_eeprom(link_fd))
	{
		disp_label ("Unable to load eeprom data");
		exit (1);
	}
	if(get_all_nodes(link_fd))
	{
		disp_label ("Unable to find node info");
		return (1);
	}
#endif

#if 0
	/* Retrieve configuration data */
	disp_label ("Loading Sub-system configuration data");
	if (!(Tconfigp_ptr = (CONFIGURATION *) malloc (sizeof (CONFIGURATION))))
	{
		disp_label ("Unable to allocate memory for configuration data");
		return (1);
	}
	memset (Tconfigp_ptr, 0, sizeof (CONFIGURATION));
	if (config_load (Tconfigp_ptr) < 0)
	{
		disp_label ("Unable to load configuration data");
		return (1);
	}
#endif


	while (config_ptr == NULL)
	{
		struct timeval to;
		int I_rc;
//		int	maxfd;
		fd_set readfds;

		to.tv_sec = 0;
		to.tv_usec = 5000;
		FD_ZERO (&readfds);
//		FD_SET (I_sockfd, &readfds);
		FD_SET (link_fd, &readfds);
//		maxfd = MAX (I_sockfd, link_fd) + 1;

//		if ((I_rc = select (maxfd, &readfds, 0, 0, &to)) < 0)
		if ((I_rc = select (link_fd+1, &readfds, 0, 0, &to)) < 0)
		{
			if (errno = EINTR)
				continue;								/* We were interrupted by a signal... continue james... */
			perror ("select");
			return (1);
		}
		else if (!I_rc)
		{
			update_spinner();
			continue;
		}
		else if (FD_ISSET (link_fd, &readfds))
		{
			/* process link data */
			if (node_link (link_fd, 0, 0) < 0)
			{
				perror ("bad facility link");
				return (1);
			}
		}
		else
		{
			disp_label ("select() assert! failure");
			return (1);
		}
	}
	return(0);
}


/*********************************************************************/

disp_label (char *msg)
{
	printf ("[%s] %s\n", mytime (), msg);
}

/*********************************************************************/

/* Debug thing for timestamps */
char timebuff[64];
char *
mytime ()
{
	time_t timeval;

	timeval = time (NULL);
	strcpy (timebuff, ctime (&timeval));
	timebuff[(strlen (timebuff) - 1)] = 0;
	return (timebuff);
}

/*********************************************************************/

int I_last = 0;
char spinnything[] =
{'|', '/', '-', '\\'};
int 
update_spinner ()
{

	printf ("%c%c", spinnything[I_last], (char) 8);
	fflush (stdout);
	I_last++;
	if (I_last == 4)
		I_last = 0;
}

/*********************************************************************/

int	checksum(char *packet)
{
	return(0);
}

/*********************************************************************/

int	build_checksum(char *packet)
{
	return(0xABCD);
}

/*********************************************************************/

/*
	This routine is normally called from the select() semaphore.
	In such case it will read whatever is waiting in COM2 port, and if the
	data can be assembled into a packet - process it.

	Another way of calling it is from a wait_reply() routine.
	In this case it may have no data - return REPLY_NO_DATA


	Return: -3 == no data
					-2 == requested packet ready
					
					len== number of bytes in the buffer
						0== processed OK
*/

int node_link (int link_fd, int opcode, char* buff)
{
static char packet_buf[128];
static int pack_size;
static unsigned int pack_time;

char packet[128];
char reply[16];
char *cur_packet;
int	pack_err;
int pack_code;
int i,j,k;


printf("into node_link...\n");

	i = read (link_fd, packet, 128);

#if 1
	deb(printf("node_link[%d]: ",i));
	for (j = 0; j < i; j++)
	{
		deb(printf("%02X ",packet[j]));
	}
	deb(printf("\n"));
#endif

	/* by default - try to read from the in buff. */
	cur_packet = packet;

	/* we have that many bytes waiting...*/
	if (0 != pack_size)
	{
		j = (int) time (NULL);
		if ((unsigned int)j > pack_time + STALE_PACKET_TO)
		{
			/* invalidate old bytes waiting - most likely garbage... */
			deb(printf("node_link(): stale packet(%d byte(s))...\n", pack_size));
			pack_size = 0;
		}
		else
		{
			deb(printf("node_link(): append to packet[%d]...\n", pack_size));
			cur_packet = packet_buf;


			for (k = 0; k < i; k++)
			{
				cur_packet[pack_size++] = packet[k];
			}

			i = pack_size;		//fake receiving that many bytes
		}
	}


#if 0
	/* sanity check: we do not expect all 128 bytes back, and pack size needs
		to be in range as well	*/
	if ((i == 128) || cur_packet[0] > 127)
	{
		/* fake some garbage short packet with a perfectly invalid checksum */
		cur_packet[0] = 0x04;
		cur_packet[1] = 0xFF;
		cur_packet[2] = 0x00;
		cur_packet[3] = 0x00;
	}
#endif



	if (i < cur_packet[0])	//did we receive full packet?
	{
		if(cur_packet != packet_buf)
		{
			for (j = 0; j < i; j++)
			{
				packet_buf[j] = cur_packet[j];
			}
		pack_size = i;
		pack_time = (int) time (NULL);
		}

		return(pack_size);				//no - wait for more...
	}

	/* we have the packet ready! Clear the incomplete packet ptr.*/
	pack_size = 0;


	/* else it is unsolicited message from the MPU, so process and reply */
	{
		switch (*(cur_packet+1))	//work with the command code...
		{

			case ARE_U_THERE:
				if ((*cur_packet != 0x06) || (checksum(cur_packet)))
				{
					pack_err = BAD_LINK_PACKET;
					break;
				}
				reply[0] = 0x06;

				reply[1] = 0xAA;		//pack_code
				reply[2] = 0x00;		//pack_err == no error

				reply[3] = packet[0x02] ^ packet[0x03];

				k = build_checksum(reply);
				reply[4] = k & 0xFF;
				reply[5] = k >> 8;

				pack_err = 0;
				break;


			case NODE_CONFIG:
				if (checksum(cur_packet))
				{
					pack_code = 0xFE;
					pack_err = BAD_LINK_PACKET;
					break;
				}

				/* save the node info... */
				config_ptr = packet_buf;	//fake - do not use it! - just to <>0
#if 1
				deb(printf("Node Config:\nNode: %d\nprotocol: %d\nip: %d.%d.%d.%d/%d\nkernel loc: %d.%d.%d.%d\nkernel name: %s.\n",
					cur_packet[2],
					cur_packet[3],
					cur_packet[7],
					cur_packet[6],
					cur_packet[5],
					cur_packet[4],
					cur_packet[8],
					cur_packet[12],
					cur_packet[11],
					cur_packet[10],
					cur_packet[9],
					cur_packet+13 ));
#endif



				reply[0] = 0x06;

				reply[1] = 0xAB;		//pack_code
				reply[2] = 0x00;		//pack_err == no error

				reply[3] = cur_packet[2];			//node number!!!!

				k = build_checksum(reply);
				reply[4] = k & 0xFF;
				reply[5] = k >> 8;

				pack_err = 0;

				break;


			default:
				(printf("node_link(): unexpected packet (cmd = 0x%X).\n",
				 cur_packet[1]));
				pack_code = 0xFF;
				pack_err = BAD_LINK_PACKET;
				break;
		}


		if (pack_err)
		{
				deb(printf("node_link(): returning error for cmd = 0x%X.\n", cur_packet[1]));

				reply[0] = 0x05;
				reply[1] = pack_code;
				reply[2] = pack_err;

				k = build_checksum(reply);
				reply[3] = (char)(k & 0xFF);
				reply[4] = (char)(k >> 8);
		}

		/* send the reply */
		j = write (link_fd, reply, reply[0]);

		return (0);

	}
}
