/* File: config.c
 * Desc: Code for dealing with memory based configuration data..
 * Auth: San Mehat (mods by Mark Knox)
 * Note:
 */
#include <stdio.h>
#include <config.h>
#include <fcntl.h>
#include <string.h>
#include <netinet/in.h>
char *mytime();

int config_load(CONFIGURATION *Tconfigp_ptr)
{
	FILE 							*Tfilep_ptr;
	int 							I_rc;
	char 							SZ_line[512];
	char 							*bword;
	char 							SZ_copy[512];
	int 							I_line=0;
	CONFIG_PARTITION					*Tpartition_new=0;
	FILE						 	*Tfilep_sumpipe;
	int  							I_prisum = 0,I_secsum = 0;
	char							SZ_filename[255];

	if (!(Tfilep_sumpipe=popen("/usr/bin/sum /tmp/zaphod.cfg1","r")))
		{
		perror("popen");
		return -1;
		}
	fscanf(Tfilep_sumpipe,"%d",&I_prisum);
	pclose(Tfilep_sumpipe);

	if (!(Tfilep_sumpipe=popen("/usr/bin/sum /tmp/zaphod.cfg2","r")))
		{
		perror("popen");
		return -1;
		}
	fscanf(Tfilep_sumpipe,"%d",&I_secsum);
	pclose(Tfilep_sumpipe);

	if (I_prisum!=I_secsum)
		{
		printf("[%s] Warning: Zaphod configuration inconsistancy (pri=%d, sec=%d)\n",
			mytime(),
			I_prisum,
			I_secsum);
		printf("[%s] Using backup configuration data\n",mytime());
		strcpy(SZ_filename,"/tmp/zaphod.cfg2");
		}
	else
		{
		printf("[%s] Configuration data integrity verified\n",mytime());
		strcpy(SZ_filename,"/tmp/zaphod.cfg1");
		}

	if (!(Tfilep_ptr=fopen(SZ_filename,"r")))
		{
		perror("fopen");
		printf("[%s] Unable to open %s\n",mytime(),SZ_filename);
		return -1;
		}

	while (fgets(SZ_line,sizeof(SZ_line),Tfilep_ptr))
	{
		I_line++;
		SZ_line[strlen(SZ_line)-1]=0;
//		printf("[config_load] Processing <%s>\n",SZ_line);
		strcpy(SZ_copy,SZ_line);
		if ((SZ_line[0]=='#')||(SZ_line[0]=='\n')||(SZ_line[0]==0))
			continue;
		if (!(bword=strtok(SZ_line,"=")))
			goto syntax_error;
		if (!strcmp(bword,"ZEA"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_ext_ipaddr=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"ZEM"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_ext_netmask=inet_addr(bword))<0)
				goto bad_addr;
		}
    else if (!strcmp(bword,"ZIA"))
    {
      if (Tpartition_new)
        goto yes_partition;
      if (!(bword=strtok(NULL,"=")))
        goto bad_addr;
      if ((Tconfigp_ptr->UI_int_ipaddr=inet_addr(bword))<0)
        goto bad_addr;
    }
    else if (!strcmp(bword,"ZIM"))
		{
      if (Tpartition_new)
        goto yes_partition;
      if (!(bword=strtok(NULL,"=")))
        goto bad_addr;
      if ((Tconfigp_ptr->UI_int_netmask=inet_addr(bword))<0)
        goto bad_addr;
    }
		else if (!strcmp(bword,"RO"))
		{
			int I_c1;
			char SZ_dest[32];
			char SZ_gw[32];
			char SZ_type[5];
			struct sockaddr_in *sock;
			struct sockaddr_in *socktmp;
			struct rtentry *routep;

			if (!Tpartition_new)
			{
				/* Find first empty route position in the global routes list */
				for (I_c1=0;Tconfigp_ptr->Srtentry_route[I_c1].rt_metric!=0;I_c1++);
				if (I_c1==6)
				{
					printf("[config_load] Routing table entries exhausted\n");
					goto out;
				}
				routep=&Tconfigp_ptr->Srtentry_route[I_c1];
			}
			else
			{
				routep=&Tpartition_new->Srtentry_route;
			}
			if (!(bword=strtok(NULL,":")))
				goto bad_addr;
			strcpy(SZ_dest,bword);
			if (!(bword=strtok(NULL,":")))
				goto bad_addr;
			strcpy(SZ_gw,bword);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			strcpy(SZ_type,bword);
			if (SZ_type[0]=='h')
				routep->rt_flags=(RTF_UP | RTF_HOST);
			else if (SZ_type[0]=='n')
				routep->rt_flags=(RTF_UP | RTF_GATEWAY);
			else
			{
				printf("[config_load] Routing entry type invalid (%s)\n",SZ_type);
				goto out;
			}

			/* Fill in the route destination */
			sock = (struct sockaddr_in *) &routep->rt_dst;
			sock->sin_family=AF_INET;
			sock->sin_port=0;
			if ((sock->sin_addr.s_addr=inet_addr(SZ_dest))<0)
				goto bad_addr;

			/* Now the route gateway */
			sock = (struct sockaddr_in *) &routep->rt_gateway;
			sock->sin_family=AF_INET;
			sock->sin_port=0;
			if ((sock->sin_addr.s_addr=inet_addr(SZ_gw))<0)
				goto bad_addr;

			/* Now the route netmask/genmask thingy */
			sock = (struct sockaddr_in *) &routep->rt_genmask;
			sock->sin_family=AF_INET;
			sock->sin_port=0;
			socktmp= (struct sockaddr_in *) &Tconfigp_ptr->Srtentry_route[I_c1].rt_dst;
			if (socktmp->sin_addr.s_addr==0)
				sock->sin_addr.s_addr=0;
			else
			if (!(sock->sin_addr.s_addr=Tconfigp_ptr->UI_int_netmask))
			{
				printf("[config_load] Syntax error on line %d\n",I_line);
				printf("[config_load] RO (route) specified before ZM (ZIB netmask)\n");
				goto out;
			}
			routep->rt_metric=1;
		}
		else if (!strcmp(bword,"CE"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[0]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[1]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[2]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[3]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[4]= (char) strtol(bword,(char **)NULL,16);
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			Tconfigp_ptr->UC_ctrletheraddr[5]= (char) strtol(bword,(char **)NULL,16);
		}
		else if (!strcmp(bword,"SY"))
		{
/*
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL," ")))
				goto syntax_error;
			if (!(Tconfigp_ptr->UI_syncmngrscanfreq=atoi(bword)))
				goto syntax_error;
*/
		}
		else if (!strcmp(bword,"CT"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tconfigp_ptr->UI_cardsvcsscanfreq=atoi(bword)))
				goto syntax_error;
		}
		else if (!strcmp(bword,"SL"))
		{
/*
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tconfigp_ptr->UC_slavememthrottle=(char)atoi(bword)))
				goto syntax_error;
*/
		}
		else if (!strcmp(bword,"GLB"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			strcpy(Tconfigp_ptr->SZ_globalloadbalancer,bword);
		}	
		else if (!strcmp(bword,"TZ"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"*"))) // '*' should never appear in a filepath
				goto syntax_error;
			strcpy(Tconfigp_ptr->SZ_timezone,bword);
		}			
		else if (!strcmp(bword,"NS"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tconfigp_ptr->UI_nfsserver=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"NB"))
		{
			if (Tpartition_new)
				goto yes_partition;
			if (!(bword=strtok(NULL,"*"))) // '*' should never appear in a filepath
				goto syntax_error;
			strcpy(Tconfigp_ptr->SZ_binarypath,bword);
		}			
		else if (!strcmp(bword,"PB"))
		{
			if (Tpartition_new)
			{
				printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
				printf("[config_load] PB specified when allready in partition\n");
				goto out;
			}
			else
			{
				/* New partition definition */
				if (!Tconfigp_ptr->Tpartition_root)
				{
					/* First partition */
					Tpartition_new= (CONFIG_PARTITION *) malloc (sizeof(CONFIG_PARTITION));
					memset(Tpartition_new,0,sizeof(CONFIG_PARTITION));
					Tconfigp_ptr->Tpartition_root=Tpartition_new;
				}
				else
				{
					CONFIG_PARTITION* Tpartitionp_ptr;
					/* Additional partition */
					for (Tpartitionp_ptr=Tconfigp_ptr->Tpartition_root;
						Tpartitionp_ptr->next;
						Tpartitionp_ptr=Tpartitionp_ptr->next);
					Tpartition_new= (CONFIG_PARTITION *) malloc (sizeof(CONFIG_PARTITION));
					memset(Tpartition_new,0,sizeof(CONFIG_PARTITION));
					Tpartitionp_ptr->next = Tpartition_new;
				}
			/* Partition structure linked in */
			}
		}
		else if (!strcmp(bword,"AI"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tpartition_new->UI_aliasaddr=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"AM"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"=")))
				goto bad_addr;
			if ((Tpartition_new->UI_aliasmask=inet_addr(bword))<0)
				goto bad_addr;
		}
		else if (!strcmp(bword,"PO"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tpartition_new->US_portbinding=htons((unsigned short)atoi(bword))))
				goto syntax_error;
		}
		else if (!strcmp(bword,"OF"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tpartition_new->C_optionflags=(char)atoi(bword)))
				goto syntax_error;
		}
		else if (!strcmp(bword,"NM"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in name */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_servername,bword);
		}
		else if (!strcmp(bword,"DR"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in a filepath */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_documentpath,bword);
		}
		else if (!strcmp(bword,"CR"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in a filepath */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_cgipath,bword);
		}
		else if (!strcmp(bword,"LB"))
		{
			if (!Tpartition_new)
				goto no_partition;
			if (!(bword=strtok(NULL,"*"))) /* * should never appear in a filepath */
				goto syntax_error;
			strcpy(Tpartition_new->SZ_loadbalancer,bword);
		}
		else if (!strcmp(bword,"NO"))
		{
			CONFIG_NODELIST *Tnodelistp_new;
			CONFIG_NODELIST *Tnodelistp_walk;

			if (!Tpartition_new)
				goto no_partition;

			if (!Tpartition_new->Tnodelist_root)
			{
				Tnodelistp_new = (CONFIG_NODELIST *) malloc(sizeof(CONFIG_NODELIST));
				memset(Tnodelistp_new,0,sizeof(CONFIG_NODELIST));
				Tpartition_new->Tnodelist_root=Tnodelistp_new;
			}
			else
			{
				for (Tnodelistp_walk=Tpartition_new->Tnodelist_root;
					Tnodelistp_walk->next;
					Tnodelistp_walk=Tnodelistp_walk->next);
				Tnodelistp_new= (CONFIG_NODELIST *) malloc (sizeof(CONFIG_NODELIST));
				memset(Tnodelistp_new,0,sizeof(CONFIG_NODELIST));
				Tnodelistp_walk->next=Tnodelistp_new;
			}

			/* Created and linked... now populate the sucker */
			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tnodelistp_new->I_nodenumber=(unsigned int)atoi(bword)))
				goto syntax_error;

			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tnodelistp_new->UI_throughputchoke=(unsigned int)atoi(bword)))
				goto syntax_error;

			if (!(bword=strtok(NULL,":")))
				goto syntax_error;
			if (!(Tnodelistp_new->UC_priority=(unsigned char)atoi(bword)))
				goto syntax_error;
		}
		else if (!strcmp(bword,"PE"))
		{
			if (!Tpartition_new)
			{
				printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
				printf("[config_load] PE specified when not in partition\n");
				goto out;
			}
			Tpartition_new=0;
		}
		else
		{
			goto syntax_error;
		}
	}

	fclose(Tfilep_ptr);

	return(0);

syntax_error:
	printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
	printf("[config_load] (%s)\n",SZ_copy);
	fclose(Tfilep_ptr);
	return(-1);

bad_addr:
	printf("[config_load] Bad inet address on line %d of configuration file\n",I_line);
	printf("[config_load] (%s)\n",SZ_copy);
	fclose(Tfilep_ptr);
	return(-1);
out:
	fclose(Tfilep_ptr);
	return(-1);
no_partition: 
	printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
	printf("[config_load] Partition token detected not within partition\n");
	fclose(Tfilep_ptr);
	return(-1);
yes_partition: 
	printf("[config_load] Syntax Error on line %d of configuration file\n",I_line);
	printf("[config_load] Non-Partition token detected within partition\n");
	fclose(Tfilep_ptr);
	return(-1);
}

destroy_partitions(CONFIGURATION *Tconfigp_ptr)
{
	CONFIG_PARTITION *Tpartitionp_begin=Tconfigp_ptr->Tpartition_root;
	CONFIG_PARTITION *Tpartitionp_ptr;
	CONFIG_NODELIST  *Tnodelistp_begin;
	CONFIG_NODELIST  *Tnodelistp_ptr;

	
	while(Tpartitionp_begin) /* While there are partitions in this config */
		{
		Tnodelistp_begin=Tpartitionp_begin->Tnodelist_root;
		while(Tnodelistp_begin) /* While there are nodes in the nodelist */
			{
			Tpartitionp_begin->Tnodelist_root=Tnodelistp_begin->next; /* Short circuit */
			free (Tnodelistp_begin);
			Tnodelistp_begin=Tpartitionp_begin->Tnodelist_root; /* reset */
			}
		Tconfigp_ptr->Tpartition_root = Tpartitionp_begin->next; /* short circuit */
		free (Tpartitionp_begin);
		Tpartitionp_begin=Tconfigp_ptr->Tpartition_root;
		}

}
